/* UART Buffer Defines */
#define USART_RX_BUFFER_SIZE 128     /* 2,4,8,16,32,64,128 or 256 bytes */
#define USART_TX_BUFFER_SIZE 16     /* 2,4,8,16,32,64,128 or 256 bytes */
#define USART_RX_BUFFER_MASK ( USART_RX_BUFFER_SIZE - 1 )
#define USART_TX_BUFFER_MASK ( USART_TX_BUFFER_SIZE - 1 )
#if ( USART_RX_BUFFER_SIZE & USART_RX_BUFFER_MASK )
	#error RX buffer size is not a power of 2
#endif
#if ( USART_TX_BUFFER_SIZE & USART_TX_BUFFER_MASK )
	#error TX buffer size is not a power of 2
#endif

/* Static Variables */
static unsigned char USART_RxBuf[USART_RX_BUFFER_SIZE];
static volatile unsigned char USART_RxHead;
static volatile unsigned char USART_RxTail;
static unsigned char USART_TxBuf[USART_TX_BUFFER_SIZE];
static volatile unsigned char USART_TxHead;
static volatile unsigned char USART_TxTail;

unsigned char USART_RxSize,USART_TxSize;

/* Prototypes */
void usart_init( unsigned int baudrate );
unsigned char usart_getc( void );
void usart_putc( unsigned char data );
unsigned char usart_unread_data(void);
unsigned char usart_unsent_data(void);
void usart_flushRx();
void usart_flushTx();
unsigned char usart_wait_for_data(unsigned char num_bytes);

/* Initialize USART */
void usart_init( unsigned int baudrate )
{
	/* Set the baud rate */
	UBRRH = (unsigned char) (baudrate>>8);                  
	UBRRL = (unsigned char) baudrate;
	
	/* Enable UART receiver and transmitter */
	UCSRB = ( ( 1 << RXCIE ) | ( 1 << RXEN ) | ( 1 << TXEN ) ); 
	
	/* Set frame format: 8 data 2stop */
	//UCSR0C = (1<<USBS0)|(1<<UCSZ01)|(1<<UCSZ00);              //For devices with Extended IO
	UCSRC = (1<<URSEL)|(1<<USBS)|(1<<UCSZ1)|(1<<UCSZ0);   //For devices without Extended IO    

	usart_flushRx();
	usart_flushTx();
}

/* Interrupt handlers */
#pragma vector=USART_RXC_vect 
__interrupt void USART_RX_interrupt( void )
{
	unsigned char data;
	unsigned char tmphead;

	/* Read the received data */
	data = UDR;                 
	/* Calculate buffer index */
	tmphead = ( USART_RxHead + 1 ) & USART_RX_BUFFER_MASK;
	USART_RxHead = tmphead;      /* Store new index */
        USART_RxSize++;
	if ( tmphead == USART_RxTail )
	{
          usart_flushRx();
	}
	
	USART_RxBuf[tmphead] = data; /* Store received data in buffer */
}

#pragma vector=USART_UDRE_vect 
__interrupt void USART_TX_interrupt( void )
{
	unsigned char tmptail;

	/* Check if all data is transmitted */
	if ( USART_TxHead != USART_TxTail )
	{
		/* Calculate buffer index */
		tmptail = ( USART_TxTail + 1 ) & USART_TX_BUFFER_MASK;
		USART_TxTail = tmptail;      /* Store new index */
	        USART_TxSize--;
	
		UDR = USART_TxBuf[tmptail];  /* Start transmition */
	}
	else
	{
		UCSRB &= ~(1<<UDRIE);         /* Disable UDRE interrupt */
	}
}

/* Read and write functions */
unsigned char usart_getc( void )
{
	unsigned char tmptail;
	
	while ( USART_RxHead == USART_RxTail )  /* Wait for incomming data */
		;
	tmptail = ( USART_RxTail + 1 ) & USART_RX_BUFFER_MASK;/* Calculate buffer index */
	
	USART_RxSize--;
	
	USART_RxTail = tmptail;                /* Store new index */
	
	return USART_RxBuf[tmptail];           /* Return data */
}

void usart_putc( unsigned char data )
{
	unsigned char tmphead;
	/* Calculate buffer index */
	tmphead = ( USART_TxHead + 1 ) & USART_TX_BUFFER_MASK; /* Wait for free space in buffer */
	while ( tmphead == USART_TxTail );

	USART_TxBuf[tmphead] = data;           /* Store data in buffer */
	USART_TxHead = tmphead;                /* Store new index */
        USART_TxSize++;

	UCSRB |= (1<<UDRIE);                    /* Enable UDRE interrupt */
}

unsigned char DataInReceiveBuffer( void )
{
	return ( USART_RxHead != USART_RxTail ); /* Return 0 (FALSE) if the receive buffer is empty */
}

unsigned char usart_unread_data(void) {
  return USART_RxSize;
}

unsigned char usart_unsent_data(void) {
  return USART_TxSize;
}

void usart_flushRx()
{
 USART_RxHead=0;
 USART_RxTail=0;
 USART_RxSize=0;
}

void usart_flushTx()
{
 USART_TxHead=0;
 USART_TxTail=0;
 USART_TxSize=0;
}

unsigned char usart_wait_for_data(unsigned char num_bytes)
{
 //waits for num_bytes bytes, but times out if they don't come
 unsigned char f=0,g=0;
 while(USART_RxSize<num_bytes)
 {
   g++;
   if(g==0xFF)
   {
     f++;
     g=0;
   }
   if(f==0xFF)
    return 0;
 }
 return 1;
}
